import inspect
from functools import partial

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http import HttpRequest
from django.test import TestCase, RequestFactory
from django.views.decorators.csrf import csrf_exempt
from rest_framework.authentication import SessionAuthentication
from rest_framework.request import Request
from rest_framework.views import APIView

from django_alt.endpoints import Endpoint as DefaultEndpoint, _KW_CONFIG_URL_FIELDS, MetaEndpoint, ViewPrototype
from django_alt.serializers import ValidatedSerializer, ValidatedModelSerializer
from rest_framework import serializers
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.utils.serializer_helpers import ReturnList

from django_alt.validators import Validator
from django_alt_tests.conf.models import ModelA


class Endpoint(DefaultEndpoint):
    custom = True

    @classmethod
    def as_view(cls):
        """
        Necessary to overcome DRF updates
        """
        if not len(cls.config):
            raise ImproperlyConfigured(('You are trying to call `as_view` on an endpoint with an empty config.\n'
                                        'Did you forget to specify the `config` attribute?\n'
                                        'Offending endpoint: `{0}`'
                                        ).format(cls.__qualname__))
        view_func = partial(ViewPrototype.respond, cls())
        view_func.__name__, view_func.__module__, view_func.__doc__ = cls.__name__, cls.__module__, cls.__doc__
        return view_func


class ConcreteSerializer(ValidatedSerializer):
    class Meta:
        model = ModelA


class ConcreteValidator(Validator):
    pass


class ConcreteModelSerializer(ValidatedModelSerializer):
    class Meta:
        model = ModelA
        validator = ConcreteValidator
        fields = '__all__'


class ConcreteEndpoint(Endpoint):
    serializer = ConcreteSerializer
    config = {
        'get': None
    }


class EndpointBaseViewLogicTests(TestCase):
    def setUp(self):
        class EBase(Endpoint):
            serializer = ConcreteSerializer
            config = {'get': None}

        self.EBase = EBase
        ModelA.objects.create(field_1='a', field_2=1)
        ModelA.objects.create(field_1='b', field_2=2)

    def patchE(self, method, is_anonymous=False):
        self.EBase.on_get = method
        return self.EBase.as_view()(self.mock_request('GET', is_anonymous))

    def mock_request(self, method, is_anonymous=False, data=None, query_params=None):
        class MockedHttpRequest(HttpRequest):
            def __init__(self):
                super().__init__()
                self.method = method
                self.user = type('UserMock', tuple(), dict(is_anonymous=is_anonymous, is_active=True))
                self.data = data
                self.encoding = 'utf8'
                self.GET = query_params or {}

        return MockedHttpRequest()

    def test_should_return_405_for_nonexistent_method_positive(self):
        req = self.mock_request('POST')
        resp = ConcreteEndpoint.as_view()(req)
        self.assertEqual(resp.status_code, 405)

    def test_overridden_endpoint_handler_return_type_checks_negative(self):
        def on_get_1(self, context, **url):
            return None

        def on_get_2(self, context, **url):
            return 200

        def on_get_3(self, context, **url):
            return {'a', 'b'}

        def on_get_4(self, context, **url):
            return {'a': 'b'}

        def on_get_5(self, context, **url):
            return {'a': 'b'}, 200

        def on_get_6(self, context, **url):
            return Response({'a': 'b'}, 200)

        def on_get_7(self, context, **url):
            return {'a': 'b'}, 200, 'foo'

        def on_get_8(self, context, **url):
            return {'a': 'b'}, 'foo'

        def on_get_9(self, context, **url):
            return [1, 2, 3]

        def on_get_10(self, context, **url):
            return ReturnList([1, 2, 3], serializer=None)

        for method in (on_get_1, on_get_3, on_get_7, on_get_8):
            with self.assertRaises(AssertionError):
                self.patchE(method)

        all((self.patchE(on_get_4), self.patchE(on_get_5),
             self.patchE(on_get_6), self.patchE(on_get_9), self.patchE(on_get_10)))

    def test_overridden_endpoint_handler_handles_rest_validation_error_positive(self):
        def on_get(self, context, **url):
            raise serializers.ValidationError({'err': ['msg']})

        self.assertEqual(self.patchE(on_get).status_code, 400)

    def test_overridden_endpoint_handler_handles_django_validation_error_positive(self):
        def on_get(self, context, **url):
            raise ValidationError({'err': ['msg']})

        self.assertEqual(self.patchE(on_get).status_code, 400)

    def test_overridden_endpoint_handler_handles_permission_error_not_logged_in_positive(self):
        def on_get(self, context, **url):
            raise PermissionError()

        self.assertEqual(self.patchE(on_get, True).status_code, 401)

    def test_overridden_endpoint_handler_handles_permission_error_logged_in_positive(self):
        def on_get(self, context, **url):
            raise PermissionError()

        self.assertEqual(self.patchE(on_get, False).status_code, 403)

    def test_overridden_endpoint_handler_handles_django_http404_positive(self):
        def on_get(self, context, **url):
            get_object_or_404(ModelA, field_1__exact='nonexistent field')

        self.assertEqual(self.patchE(on_get).status_code, 404)
        self.assertEqual(self.patchE(on_get).data, {'non_field_errors': ['No ModelA matches the given query.']})

    def test_overridden_endpoint_handler_handles_object_does_not_exist_positive(self):
        def on_get(self, context, **url):
            ModelA.objects.get(field_1__exact='nonexistent field')

        self.assertEqual(self.patchE(on_get).status_code, 404)
        self.assertEqual(self.patchE(on_get).data, {'non_field_errors': ['ModelA matching query does not exist.']})

    """
    GET handlers
    """

    def test_default_get_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {'get': None}

        response = ConcreteEndpoint.as_view()(self.mock_request('GET'))
        self.assertDictEqual(response.data, {'field_2': None, 'field_1': ''})
        self.assertEqual(response.status_code, 200)

    def test_default_get_single_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'get': {
                    'query': lambda model, **url: model.objects.first()
                }
            }

        response = ConcreteEndpoint.as_view()(self.mock_request('GET'))
        self.assertDictEqual(response.data, {'id': 1, 'field_2': 1, 'field_1': 'a'})
        self.assertEqual(response.status_code, 200)

    def test_default_get_many_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'get': {
                    'query': lambda model, **url: model.objects.all()
                }
            }

        response = ConcreteEndpoint.as_view()(self.mock_request('GET'))
        self.assertListEqual(response.data,
                             [{'id': 1, 'field_2': 1, 'field_1': 'a'}, {'id': 2, 'field_2': 2, 'field_1': 'b'}])
        self.assertEqual(response.status_code, 200)

    def test_override_get_endpoint_handler_positive(self):
        class OverriddenEndpoint(Endpoint):
            serializer = ConcreteSerializer
            config = {'get': None}

            def on_get(self, context, **url):
                return {'test': 'ok'}, 201

        resp = OverriddenEndpoint.as_view()(self.mock_request('GET'))
        self.assertEqual(resp.status_code, 201)
        self.assertDictEqual(resp.data, {'test': 'ok'})

    """
    POST handlers
    """

    def test_default_post_none_and_data_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'post': None
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('POST'))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'non_field_errors': ['No data provided']})

    def test_default_post_none_and_data_empty_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'post': None
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('POST', data={}))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data,
                             {'field_2': ['This field is required.'], 'field_1': ['This field is required.']})

    def test_default_post_single_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'post': None
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('POST', data={'field_2': 1, 'field_1': 'a'}))
        self.assertEqual(resp.status_code, 201)
        self.assertDictEqual(resp.data, {'id': 3, 'field_2': 1, 'field_1': 'a'})
        self.assertEqual(ModelA.objects.count(), 3)
        self.assertEqual(ModelA.objects.last().id, 3)
        self.assertEqual(ModelA.objects.last().field_2, 1)
        self.assertEqual(ModelA.objects.last().field_1, 'a')

    def test_default_post_many_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'post': None
            }

        resp = ConcreteEndpoint.as_view()(
            self.mock_request('POST', data=[{'field_2': 1, 'field_1': 'a'}, {'field_2': 2, 'field_1': 'b'}]))
        self.assertEqual(resp.status_code, 201)
        self.assertListEqual(resp.data,
                             [{'id': 3, 'field_2': 1, 'field_1': 'a'}, {'id': 4, 'field_2': 2, 'field_1': 'b'}])
        self.assertEqual(ModelA.objects.count(), 4)
        self.assertEqual(ModelA.objects.all()[2].id, 3)
        self.assertEqual(ModelA.objects.all()[2].field_2, 1)
        self.assertEqual(ModelA.objects.all()[2].field_1, 'a')
        self.assertEqual(ModelA.objects.all()[3].id, 4)
        self.assertEqual(ModelA.objects.all()[3].field_2, 2)
        self.assertEqual(ModelA.objects.all()[3].field_1, 'b')

    def test_default_post_many_not_allowed_handler_negative(self):
        with self.assertRaises(AssertionError):
            class ConcreteEndpoint(Endpoint):
                serializer = ConcreteModelSerializer
                config = {
                    'post': None,
                    'allow_many': False
                }

            ConcreteEndpoint.as_view()(
                self.mock_request('POST', data=[{'field_2': 1, 'field_1': 'a'}, {'field_2': 2, 'field_1': 'b'}]))

    """
    PATCH handlers
    """

    def test_default_patch_query_returns_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda *args, **kwargs: None
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH'))
        self.assertEqual(resp.status_code, 404)

    def test_default_patch_query_raises_no_data_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH'))
        self.assertEqual(resp.status_code, 404)
        self.assertDictEqual(resp.data, {'non_field_errors': ['ModelA matching query does not exist.']})

    def test_default_patch_query_raises_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH', data={'field_2': 5}))
        self.assertEqual(resp.status_code, 404)
        self.assertDictEqual(resp.data, {'non_field_errors': ['ModelA matching query does not exist.']})

    def test_default_patch_none_and_data_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH'))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'non_field_errors': ['No data provided']})

    def test_default_patch_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH', data={'field_1': 'updated'}))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.data, {'id': 1, 'field_2': 1, 'field_1': 'updated'})
        self.assertEqual(ModelA.objects.first().field_1, 'updated')
        self.assertEqual(ModelA.objects.first().field_2, 1)
        self.assertEqual(ModelA.objects.first().id, 1)

    def test_default_patch_many_data_not_list_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.all()
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH', data={'field_1': 'updated'}))
        self.assertDictEqual(resp.data, {'non_field_errors': ['Expected a list of items but got type "dict".']})

    def test_default_patch_many_data_list_not_implemented_handler_negative(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.all()
                }
            }

        with self.assertRaises(NotImplementedError) as ex:
            ConcreteEndpoint.as_view()(self.mock_request('PATCH', data=[{'field_1': 'updated'}]))
        self.assertEqual(str(ex.exception),
                         'Serializers with many=True do not support multiple update by default, only multiple create. '
                         'For updates it is unclear how to deal with insertions and deletions. If you need to support '
                         'multiple update, use a `ListSerializer` class and override `.update()` so you can specify '
                         'the behavior exactly.')

    def test_default_patch_data_is_a_list_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'patch': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PATCH', data=[{'field_1': 'updated'}]))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'non_field_errors': ['Invalid data. Expected a dictionary, but got list.']})

    """
    DELETE handlers
    """

    def test_default_delete_query_returns_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda *args, **kwargs: None
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE'))
        self.assertEqual(resp.status_code, 404)

    def test_default_delete_query_raises_no_data_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE'))
        self.assertEqual(resp.status_code, 404)
        self.assertDictEqual(resp.data, {'non_field_errors': ['ModelA matching query does not exist.']})

    def test_default_delete_query_raises_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE', data={'field_2': 5}))
        self.assertEqual(resp.status_code, 404)
        self.assertDictEqual(resp.data, {'non_field_errors': ['ModelA matching query does not exist.']})

    def test_default_delete_none_and_data_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        count = ModelA.objects.all().count()
        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE'))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.data, {'id': None, 'field_2': 1, 'field_1': 'a'})
        self.assertEqual(ModelA.objects.all().count(), count - 1)
        self.assertEqual(ModelA.objects.first().id, 2)

    def test_default_delete_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        count = ModelA.objects.all().count()

        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE', data={'field_1': 'updated'}))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.data, {'id': None, 'field_2': 1, 'field_1': 'a'})
        self.assertEqual(ModelA.objects.all().count(), count - 1)
        self.assertEqual(ModelA.objects.first().id, 2)

    def test_default_delete_many_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'delete': {
                    'query': lambda model, **url: ModelA.objects.all()
                }
            }

        count = ModelA.objects.all().count()
        assert count > 0
        resp = ConcreteEndpoint.as_view()(self.mock_request('DELETE', data={'field_1': 'updated'}))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data,
                         [{'id': None, 'field_2': 1, 'field_1': 'a'}, {'id': None, 'field_2': 2, 'field_1': 'b'}])
        self.assertEqual(ModelA.objects.all().count(), 0)

    """
    PUT handlers
    """

    def test_default_put_query_returns_none_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda *args, **kwargs: None
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT'))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data,
                             {'field_2': ['This field is required.'], 'field_1': ['This field is required.']})

    def test_default_put_query_raises_no_data_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                    # PUT endpoint works like POST if an object is not found
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT'))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data,
                             {'field_2': ['This field is required.'], 'field_1': ['This field is required.']})

    def test_default_put_query_raises_incomplete_data_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'x'}))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'field_2': ['This field is required.']})

    def test_default_put_query_acts_as_post_queryset_raises_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        self.assertEqual(ModelA.objects.count(), 2)
        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'x', 'field_2': 5}))
        self.assertEqual(resp.status_code, 201)
        self.assertDictEqual(resp.data, {'field_1': 'x', 'field_2': 5, 'id': 3})
        self.assertEqual(ModelA.objects.count(), 3)
        self.assertEqual(ModelA.objects.last().id, 3)
        self.assertEqual(ModelA.objects.last().field_1, 'x')
        self.assertEqual(ModelA.objects.last().field_2, 5)

    def test_default_put_query_acts_as_post_queryset_empty_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.filter(field_1='nonexistent')
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'x', 'field_2': 5}))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'non_field_errors': ['Expected a list of items but got type "dict".']})

    def test_default_put_query_acts_as_post_queryset_raises_data_many_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        self.assertEqual(ModelA.objects.count(), 2)
        resp = ConcreteEndpoint.as_view()(
            self.mock_request('PUT', data=[{'field_1': 'x', 'field_2': 5}, {'field_1': 'y', 'field_2': 6}]))
        self.assertEqual(resp.status_code, 201)
        self.assertEqual(resp.data, [{'id': 3, 'field_1': 'x', 'field_2': 5}, {'id': 4, 'field_1': 'y', 'field_2': 6}])
        self.assertEqual(ModelA.objects.count(), 4)
        self.assertEqual(ModelA.objects.all()[2].id, 3)
        self.assertEqual(ModelA.objects.all()[2].field_1, 'x')
        self.assertEqual(ModelA.objects.all()[2].field_2, 5)
        self.assertEqual(ModelA.objects.all()[3].id, 4)
        self.assertEqual(ModelA.objects.all()[3].field_1, 'y')
        self.assertEqual(ModelA.objects.all()[3].field_2, 6)

    def test_default_put_query_acts_as_post_many_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.get(field_1='nonexistent')
                }
            }

        self.assertEqual(ModelA.objects.count(), 2)
        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'x', 'field_2': 5}))
        self.assertEqual(resp.status_code, 201)
        self.assertDictEqual(resp.data, {'field_1': 'x', 'field_2': 5, 'id': 3})
        self.assertEqual(ModelA.objects.count(), 3)
        self.assertEqual(ModelA.objects.last().id, 3)
        self.assertEqual(ModelA.objects.last().field_1, 'x')
        self.assertEqual(ModelA.objects.last().field_2, 5)

    def test_default_put_acts_as_patch_disallow_partial_update_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'y'}))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(resp.data, {'field_2': ['This field is required.']})

    def test_default_put_acts_as_patch_handler_positive(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.first()
                }
            }

        self.assertEqual(ModelA.objects.count(), 2)
        resp = ConcreteEndpoint.as_view()(self.mock_request('PUT', data={'field_1': 'y', 'field_2': 1337}))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.data, {'field_1': 'y', 'field_2': 1337, 'id': 1})
        self.assertEqual(ModelA.objects.count(), 2)
        self.assertEqual(ModelA.objects.first().id, 1)
        self.assertEqual(ModelA.objects.first().field_1, 'y')
        self.assertEqual(ModelA.objects.first().field_2, 1337)

    def test_default_put_acts_as_patch_many_not_implemented_handler_negative(self):
        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'put': {
                    'query': lambda model, **url: ModelA.objects.all()
                }
            }

        with self.assertRaises(NotImplementedError) as ex:
            ConcreteEndpoint.as_view()(self.mock_request('PUT', data=[{'field_1': 'y', 'field_2': 1337}]))
        self.assertEqual(str(ex.exception), 'Serializers with many=True do not support multiple update by default, '
                                            'only multiple create. For updates it is unclear how to deal with '
                                            'insertions and deletions. If you need to support multiple update, '
                                            'use a `ListSerializer` class and override `.update()` so you can specify '
                                            'the behavior exactly.')

    """
    GET handlers with filtering
    """

    def test_default_get_with_single_filter_positive(self):
        called = None

        def first_filter(objects, value):
            nonlocal called
            called = value
            return objects.first()

        def last_filter(objects, value):
            nonlocal called
            called = value
            return objects.last()

        class ConcreteEndpoint(Endpoint):
            serializer = ConcreteModelSerializer
            config = {
                'get': {
                    'query': lambda model, **url: ModelA.objects.all(),
                    'filters': {
                        'first': first_filter,
                        'last': last_filter
                    }
                }
            }

        resp = ConcreteEndpoint.as_view()(self.mock_request('GET', data=[{'field_1': 'y', 'field_2': 1337}]))
        self.assertEqual(resp.data, [{'id': 1, 'field_2': 1, 'field_1': 'a'}, {'id': 2, 'field_2': 2, 'field_1': 'b'}])

        resp = ConcreteEndpoint.as_view()(
            self.mock_request('GET', data=[{'field_1': 'y', 'field_2': 1337}], query_params={'first': 'param'}))
        self.assertDictEqual(resp.data, {'id': 1, 'field_2': 1, 'field_1': 'a'})
        self.assertEqual(called, 'param')

        resp = ConcreteEndpoint.as_view()(
            self.mock_request('GET', data=[{'field_1': 'y', 'field_2': 1337}], query_params={'last': 'param2'}))
        self.assertDictEqual(resp.data, {'id': 2, 'field_2': 2, 'field_1': 'b'})
        self.assertEqual(called, 'param2')


class EndpointDefinitionLogicTests(TestCase):
    def test_empty_definition_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint(Endpoint):
                pass
        self.assertIn('`serializer` is required', ex.exception.args[0])

    def test_empty_definition_with_custom_set_positive(self):
        class MyEndpoint(Endpoint):
            custom = True

    def test_serializer_is_arbitrary_object_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint(Endpoint):
                serializer = 'foo'
        self.assertIn('`serializer` must be a class', ex.exception.args[0])

    def test_no_model_and_serializer_no_meta_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint(Endpoint):
                serializer = Serializer
        self.assertIn('Explicit `model` field is missing', ex.exception.args[0])

    def test_config_is_not_dict(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint(Endpoint):
                serializer = ConcreteSerializer
                config = 'foo'
        self.assertIn('`config` field must be of type `dict`', ex.exception.args[0])

    def test_unknown_method_foo_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint1(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'foo': {
                        'query': None
                    }
                }
        self.assertIn('unknown HTTP method `foo`', ex.exception.args[0])

    def test_unknown_method_selector_keys_checking_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint1(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'unknown1': None,
                    }
                }
        self.assertIn('`MyEndpoint1` get config contains unknown keys: `unknown1`.', ex.exception.args[0])

    def test_many_unknown_method_selector_keys_checking_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint1(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'unknown1': None,
                        'unknown2': None
                    }
                }
        self.assertIn('`MyEndpoint1` get config contains unknown keys:', ex.exception.args[0])
        self.assertIn('unknown2', ex.exception.args[0])
        self.assertIn('unknown1', ex.exception.args[0])

    def test_query_is_not_callable_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint2(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'query': None
                    }
                }
        self.assertIn('must be a callable', ex.exception.args[0])
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint7(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'query': {'something_strange'},
                        'filters': {
                            'x': lambda x, y: (x, y)
                        }
                    }}
        self.assertIn('must be a callable', ex.exception.args[0])

    def test_method_definition_is_not_passed_an_iterable_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint2(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': lambda a: a
                }
        self.assertIn('Incorrectly formed endpoint config object', ex.exception.args[0])

    def test_get_filters_without_query_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint3(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'filters': {
                            'y': lambda x, y: (x, y)
                        }
                    }}
        self.assertIn('cannot be used without', ex.exception.args[0])

    def test_get_filters_not_dict_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint5(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'get': {
                        'query': lambda x, y: (x, y),
                        'filters': lambda x, y: (x, y)
                    }}
        self.assertIn('field must be of type `dict`', ex.exception.args[0])

    def test_delete_called_without_query_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint8(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'delete': None
                }
        self.assertIn('`config` for `delete` must include', ex.exception.args[0])

    def test_url_fields_not_an_iterable_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint8(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'post': {
                        _KW_CONFIG_URL_FIELDS: 'foo'
                    }
                }
        self.assertIn('field must be an iterable', ex.exception.args[0])

    def test_patch_includes_query_attribute_negative(self):
        with self.assertRaises(AssertionError) as ex:
            class MyEndpoint8(Endpoint):
                serializer = ConcreteSerializer
                config = {
                    'patch': {}
                }
        self.assertIn('`patch` method config must include', ex.exception.args[0])

    def test_url_fields_is_iterable_positive(self):
        class MyEndpoint8(Endpoint):
            serializer = ConcreteSerializer
            config = {'post': {_KW_CONFIG_URL_FIELDS: ('foo',)}}

        class MyEndpoint9(Endpoint):
            serializer = ConcreteSerializer
            config = {'post': {_KW_CONFIG_URL_FIELDS: ['foo', ]}}

        class MyEndpoint10(Endpoint):
            serializer = ConcreteSerializer
            config = {'post': {_KW_CONFIG_URL_FIELDS: {'foo', 'bar'}}}

    def test_allow_many_flag_set_to_true_automatically_positive(self):
        class MyEndpoint6(Endpoint):
            serializer = ConcreteSerializer
            config = {
                'get': {'query': lambda x, y: (x, y), },
                'post': {'query': lambda x, y: (x, y), },
                'put': {'query': lambda x, y: (x, y), },
                'patch': {'query': lambda x, y: (x, y), },
                'delete': {'query': lambda x, y: (x, y), },
            }

        self.assertTrue(MyEndpoint6.config['get'].allow_many)
        self.assertTrue(MyEndpoint6.config.post.allow_many)
        self.assertTrue(MyEndpoint6.config.put.allow_many)
        self.assertTrue(MyEndpoint6.config.patch.allow_many)
        self.assertTrue(MyEndpoint6.config.delete.allow_many)

    def test_get_with_filters_and_delete_with_query_positive(self):
        class MyEndpoint6(Endpoint):
            serializer = ConcreteSerializer
            config = {
                'get': {
                    'query': lambda x, y: (x, y),
                    'filters': {
                        'x': lambda x, y: (x, y)
                    }
                },
                'delete': {
                    'query': lambda x, y: (x, y)
                }}

    def test_get_with_filters_positive(self):
        class MyEndpoint4(Endpoint):
            serializer = ConcreteSerializer
            config = {
                'get': {
                    'query': lambda x, y: (x, y),
                    'filters': {
                        'x': lambda x, y: (x, y)
                    }
                }}

    def test_transform_config_shorthands_positive(self):
        def are_equal(c1, c0):
            c1 = MetaEndpoint.transform_config_shorthands('MockEndpoint', c1)
            self.assertDictEqual(c0, c1)

        are_equal({
            'get': None
        }, {
            'get': {}
        })

        are_equal({
            'get, post': None
        }, {
            'get': {},
            'post': {}
        })

        are_equal({
            ' get,post, put': None
        }, {
            'get': {},
            'post': {},
            'put': {},
        })
        are_equal({
            ' get,post': {
                'q': 1
            }
        }, {
            'get': {'q': 1},
            'post': {'q': 1},
        })
        are_equal({
            ' get,post': {
                'q': 1
            },
            'post': {
                'z': 2
            }
        }, {
            'get': {'q': 1},
            'post': {'q': 1, 'z': 2},
        })
        are_equal({
            ' get': {
                'q': 1
            },
            'get': {
                'z': 2
            }
        }, {
            'get': {'q': 1, 'z': 2},
        })
        are_equal({
            ' get, get': {
                'q': 1
            },
            'post,get': {
                'z': 4
            }
        }, {
            'get': {'q': 1, 'z': 4},
            'post': {'z': 4}
        })


class EndpointMiscTests(TestCase):
    def test_calling_as_view_with_empty_config_negative(self):
        with self.assertRaises(ImproperlyConfigured):
            class EmptyEndpoint(Endpoint):
                serializer = ConcreteSerializer

            EmptyEndpoint.as_view()

    def test_calling_endpoint_directly_raises_negative(self):
        with self.assertRaises(ImproperlyConfigured):
            ConcreteEndpoint()()

    def test_partial_application_correctness_returned_by_as_view_positive(self):
        ep = ConcreteEndpoint.as_view()
        self.assertFalse(inspect.isclass(ep.args[0]))

    def test_view_prototype_type_normalization(self):
        func = ViewPrototype.normalize_type
        self.assertEqual(func('1'), 1)
        self.assertEqual(func('1.5'), 1.5)
        self.assertEqual(func('1.5A'), '1.5A')
        self.assertEqual(func('none'), 'none')
        self.assertEqual(func('true'), True)
        self.assertEqual(func('tRUe'), True)
        self.assertEqual(func('false'), False)
        self.assertEqual(func('False'), False)
        self.assertEqual(func('Falses'), 'Falses')



from typing import Type

from django.db.models import QuerySet, Model
from django.test import TestCase
from django_alt.serializers import ValidatedModelSerializer

from django_alt.validators import Validator
from django_alt_tests.conf.models import ModelA


class ConcreteValidator(Validator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.callstack = []

    def did_create(self, instance: Type[Model]):
        self.callstack.append('did_create')

    def clean(self):
        self.callstack.append('clean')

    def will_update(self, instance: Type[Model]):
        self.callstack.append('will_update')

    def will_create(self):
        self.callstack.append('will_create')

    def base_db(self):
        self.callstack.append('base_db')

    def did_update(self, instance: Type[Model]):
        self.callstack.append('did_update')

    def will_read(self, instance: Type[Model]):
        self.callstack.append('will_read')

    def base(self):
        self.callstack.append('base')

    def will_delete(self, queryset: Type[QuerySet]):
        self.callstack.append('will_delete')

    def did_delete(self, queryset: Type[QuerySet]):
        self.callstack.append('did_delete')


class ConcreteSerializer(ValidatedModelSerializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.calls = []

    def do_create(self, **attrs):
        self.calls.append('do_create')
        return super().do_create(**attrs)

    def do_update(self, instance, **attrs):
        self.calls.append('do_update')
        return super().do_update(instance, **attrs)

    def do_delete(self, queryset, **attrs):
        self.calls.append('do_delete')
        return super().do_delete(queryset, **attrs)

    class Meta:
        model = ModelA
        validator = ConcreteValidator
        fields = '__all__'


class ValidatedModelSerializerTests(TestCase):
    def test_basic_creation_positive(self):
        serializer = ConcreteSerializer(data=dict(field_1='abc', field_2=1337))
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first(), obj)
        self.assertEqual(obj.id, 1)
        self.assertEqual(obj.field_1, 'abc')
        self.assertEqual(obj.field_2, 1337)

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_create')

        self.assertEqual(serializer.validator.callstack, ['clean', 'base', 'will_create', 'base_db', 'did_create'])

    def test_basic_reading_positive(self):
        obj = ModelA.objects.create(field_1='abc', field_2=1337)
        serializer = ConcreteSerializer(obj)
        data = serializer.data
        self.assertDictEqual(data, {
            'id': 1,
            'field_1': 'abc',
            'field_2': 1337,
        })
        self.assertEqual(serializer.validator.callstack, ['will_read'])

    def test_basic_creation_and_reading_positive(self):
        serializer = ConcreteSerializer(data=dict(field_1='abc', field_2=1337))
        serializer.is_valid(raise_exception=True)
        serializer.save()
        data = serializer.data

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_create')

        self.assertEqual(serializer.validator.callstack,
                         ['clean', 'base', 'will_create', 'base_db', 'did_create', 'will_read'])
        self.assertDictEqual(data, {
            'id': 1,
            'field_1': 'abc',
            'field_2': 1337,
        })

    def test_basic_update_positive(self):
        obj = ModelA.objects.create(field_1='abc', field_2=1337)
        serializer = ConcreteSerializer(obj, data=dict(field_1='def', field_2=43))
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_update')
        self.assertEqual(serializer.validator.callstack, ['clean', 'base', 'will_update', 'base_db', 'did_update'])
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first(), obj)
        self.assertEqual(obj.id, 1)
        self.assertEqual(obj.field_1, 'def')
        self.assertEqual(obj.field_2, 43)

    def test_basic_update_and_reading_positive(self):
        obj = ModelA.objects.create(field_1='abc', field_2=1337)
        serializer = ConcreteSerializer(obj, data=dict(field_1='def', field_2=43))
        serializer.is_valid(raise_exception=True)
        serializer.save()
        data = serializer.data

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_update')
        self.assertEqual(serializer.validator.callstack,
                         ['clean', 'base', 'will_update', 'base_db', 'did_update', 'will_read'])
        self.assertDictEqual(data, {
            'id': 1,
            'field_1': 'def',
            'field_2': 43,
        })

    def test_serializer_delete_no_validation_positive(self):
        obj = ModelA.objects.create(field_1='abc', field_2=1337)
        serializer = ConcreteSerializer(obj)

        self.assertEqual(ModelA.objects.count(), 1)
        serializer.delete()
        self.assertEqual(ModelA.objects.count(), 0)

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_delete')
        self.assertEqual(serializer.validator.callstack, ['will_delete', 'did_delete'])

        data = serializer.data
        self.assertEqual(serializer.validator.callstack, ['will_delete', 'did_delete', 'will_read'])

    def test_serializer_delete_with_validation_positive(self):
        obj = ModelA.objects.create(field_1='abc', field_2=1337)
        serializer = ConcreteSerializer(obj, data=dict(field_1='abc', field_2=1338))

        self.assertEqual(ModelA.objects.count(), 1)
        serializer.is_valid(raise_exception=True)
        serializer.delete()
        self.assertEqual(ModelA.objects.count(), 0)

        self.assertEqual(len(serializer.calls), 1)
        self.assertEqual(serializer.calls[0], 'do_delete')
        self.assertEqual(serializer.validator.callstack, ['clean', 'base', 'will_delete', 'did_delete'])

        data = serializer.data
        self.assertEqual(serializer.validator.callstack, ['clean', 'base', 'will_delete', 'did_delete', 'will_read'])



import json
from functools import partial

from rest_framework.decorators import api_view
from typing import Union, Type, Tuple

from django.core.exceptions import ValidationError as django_ValidationError, ImproperlyConfigured, ObjectDoesNotExist
from django.http.response import HttpResponseBase, Http404
from rest_framework import status
from rest_framework.fields import empty
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer, ValidationError as drf_ValidationError
from rest_framework.views import APIView

from .contexts import RequestContext
from .dotdict import ddict, undefined
from .exceptions import EndpointError
from .serializers import ValidatedSerializer
from .utils.shortcuts import invalid, make_error, first_defined, try_cast, as_bool

_HTTP_METHODS = ('get', 'post', 'patch', 'put', 'delete')

_KW_CONFIG_FILTERS = 'filters'
_KW_CONFIG_QUERYSET = 'query'
_KW_CONFIG_URL_FIELDS = 'fields_from_url'
_KW_CONFIG_DO_NOT_NORMALIZE_DATA = 'no_data_casting'
_KW_CONFIG_ALLOW_MANY = 'allow_many'
_KW_ALL = {_KW_CONFIG_FILTERS, _KW_CONFIG_QUERYSET,
           _KW_CONFIG_URL_FIELDS, _KW_CONFIG_DO_NOT_NORMALIZE_DATA,
           _KW_CONFIG_ALLOW_MANY}


class ViewPrototype:
    @staticmethod
    def normalize_type(value):
        return first_defined(
            try_cast(int, value),
            try_cast(float, value),
            as_bool(value),
            value)

    @staticmethod
    def apply_filters(qs, filters, query_params):
        current_param = None
        try:
            for param, func in filters.items():
                current_param = param
                if param in query_params:
                    qs = func(qs, query_params[param])
            return qs
        except django_ValidationError as e:
            invalid(current_param, e.message)
        except ValueError as e:
            invalid(current_param, str(e))

    @staticmethod
    def defuse_response(responder, get_context):
        context = None
        try:
            context = get_context()
            return responder(context)
        except EndpointError as e:
            return e.status_code
        except drf_ValidationError as e:
            return e.detail, status.HTTP_400_BAD_REQUEST
        except django_ValidationError as e:
            return e.message_dict, status.HTTP_400_BAD_REQUEST
        except PermissionError:
            if context is not None and context.request.user.is_anonymous:
                return status.HTTP_401_UNAUTHORIZED
            return status.HTTP_403_FORBIDDEN
        except (Http404, ObjectDoesNotExist) as e:
            return make_error(None, ', '.join(e.args)), status.HTTP_404_NOT_FOUND

    @staticmethod
    def postprocess_response(response):
        for attr in ('renderer_classes', 'parser_classes', 'authentication_classes',
                     'throttle_classes', 'permission_classes'):
            if not hasattr(response, attr):
                setattr(response, attr, getattr(APIView, attr))
        return response

    @staticmethod
    def respond(endpoint_self: Type['Endpoint'], request, *args, **kwargs):
        # TODO solve for HTTP HEAD
        method = request.method.lower()
        config = endpoint_self.config.get(method, None)
        if config is None:
            return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)

        data = getattr(request, 'data', None)
        if not config.get(_KW_CONFIG_DO_NOT_NORMALIZE_DATA, False):
            args = [ViewPrototype.normalize_type(value) for value in args]
            kwargs = {k: ViewPrototype.normalize_type(v) for k, v in kwargs.items()}
            query_params = {k: ViewPrototype.normalize_type(v) for k, v in request.GET.items()}
        else:
            query_params = {k: v for k, v in request.GET.items()}

        if _KW_CONFIG_URL_FIELDS in config:
            # TODO test this behaviour
            try:
                updated_fragment = {k: kwargs[k] for k in config[_KW_CONFIG_URL_FIELDS]}
                if isinstance(data, list):
                    for member in data:
                        assert isinstance(member, dict), (
                            'An endpoint that accepts a nested list of items\n'
                            'cannot have `{}` config set.\n'
                            'Offending endpoint: `{}`, method: `{}`.\n'
                            '`request.data` dump: \n`{}`'
                        ).format(_KW_CONFIG_URL_FIELDS, endpoint_self.__class__.__name__, request.method, data)
                        member.update(updated_fragment)
                else:
                    data.update(updated_fragment)
            except KeyError:
                raise AssertionError(('Key supplied in `{0}` was not present in the url dict at endpoint `{1}`.\n'
                                      '`{0}` dump: {2}').format(_KW_CONFIG_URL_FIELDS, endpoint_self.__class__.__name__,
                                                                json.dumps(config[_KW_CONFIG_URL_FIELDS])))

        def get_context():
            queryset = None
            try:
                if _KW_CONFIG_QUERYSET in config:
                    queryset = config[_KW_CONFIG_QUERYSET](endpoint_self.model, **kwargs)
                    if _KW_CONFIG_FILTERS in config and len(query_params):
                        queryset = ViewPrototype.apply_filters(queryset, config[_KW_CONFIG_FILTERS], query_params)
            except ObjectDoesNotExist:
                if method != 'put':
                    # put is allowed not to have a queryset
                    # an object is then created instead
                    raise
                queryset = None
            return RequestContext(request, data=endpoint_self.transform_input_data(data), queryset=queryset,
                                  url_args=args, url_kwargs=kwargs, query_params=query_params)

        responder = getattr(endpoint_self, 'on_' + method)
        result = ViewPrototype.defuse_response(responder, get_context)

        try:
            if isinstance(result, int):
                # result: int
                return ViewPrototype.postprocess_response(Response(None, result))
            elif isinstance(result, HttpResponseBase):
                # result: Response
                result.data = endpoint_self.transform_output_data(result.data) or result.data
                return ViewPrototype.postprocess_response(result)

            status_code = None
            if isinstance(result, tuple) and len(result) == 2:
                # result: Tuple[dict, int]
                assert isinstance(result[1], int)
                status_code, result = result[1], result[0]

            if isinstance(result, dict) or isinstance(result, list):
                # result: dict
                result = endpoint_self.transform_output_data(result) or result
                return ViewPrototype.postprocess_response(Response(result, status=status_code))
            raise TypeError()
        except (TypeError, AssertionError, KeyError):
            raise AssertionError(
                (
                    'Incorrect return type on method `on_{}`.\n'
                    'Endpoint methods starting with `on_` must return either:\n'
                    ' - response_data: dict\n'
                    ' - response_obj: Response\n'
                    ' - response_data, status_code: (dict, int)\n'
                    ' - status_code: int (response_data is empty)\n'
                    'Offending endpoint: `{}`.'
                ).format(method, endpoint_self.__class__.__name__))


class MetaEndpoint(type):
    def __new__(mcs, name, bases, namespace):
        if len(bases):
            # only transform for classes extending `Endpoint`
            namespace = mcs.transform_namespace(name, namespace)
        return type.__new__(mcs, name, bases, namespace)

    @classmethod
    def transform_namespace(mcs, clsname, namespace):
        skip_model_checks = namespace.get('custom', False)

        if not skip_model_checks:
            assert 'serializer' in namespace, (
                'Field `serializer` is required for an endpoint definition.\n'
                'Offending Endpoint `{}`'
            ).format(clsname)
            try:
                if not issubclass(namespace['serializer'], BaseSerializer):
                    raise AssertionError()
            except:
                raise AssertionError('Field `serializer` must be a class extending `BaseSerializer`.\n'
                                     'Offending Endpoint `{}`'.format(clsname))

        if not skip_model_checks and 'model' not in namespace:
            assert hasattr(namespace['serializer'], 'Meta') and hasattr(namespace['serializer'].Meta, 'model'), (
                'Explicit `model` field is missing and trying to get\n'
                'the `model` from the `Meta` declared on the given serializer failed\n'
                'at endpoint `{}`.'
            ).format(clsname)
            namespace['model'] = namespace['serializer'].Meta.model

        namespace['config'] = mcs.transform_config(clsname, namespace.get('config', {}), skip_model_checks)
        return namespace

    @classmethod
    def transform_config_shorthands(mcs, clsname, config):
        result = ddict()
        for k, v in config.items():
            if v is None:
                v = ddict()
            if ',' in k:
                for end in k.split(','):
                    result.setdefault(end.strip(), ddict()).update(v)
            else:
                assert hasattr(v, '__iter__'), (
                    'Incorrectly formed endpoint config object:\n'
                    '\tvalue at key `{}` must be an iterable, not `{}`.\n'
                    'Offending endpoint: `{}`, definition.'
                ).format(k.strip(), v.__class__.__name__, clsname)
                result.setdefault(k.strip(), ddict()).update(v)
        return result

    @classmethod
    def transform_config(mcs, clsname, config, skip_model_checks=False):
        assert isinstance(config, dict), (
            'Endpoint `config` field must be of type `dict`. Offending endpoint `{}`'
        ).format(clsname)

        if len(config):
            config = mcs.transform_config_shorthands(clsname, config)
            for method_name, method_config in config.items():
                assert method_name in _HTTP_METHODS, (
                    'Endpoint `{}` config definition contains an unknown HTTP method `{}`.\n'
                    'Allowed methods are `{}`.'
                ).format(clsname, method_name, _HTTP_METHODS)

                assert len(set(method_config.keys()).difference(_KW_ALL)) == 0, (
                    '`{}` {} config contains unknown keys: `{}`.'
                ).format(clsname, method_name, ', '.join(sorted(set(method_config.keys()).difference(_KW_ALL))))

                if method_name == 'patch' and not skip_model_checks:
                    assert _KW_CONFIG_QUERYSET in method_config, (
                        'Endpoint `patch` method config must include the `{}` attribute.\n'
                        'Offending endpoint: `{}`'
                    ).format(_KW_CONFIG_QUERYSET, clsname)

                method_config.setdefault(_KW_CONFIG_ALLOW_MANY, True)

                if _KW_CONFIG_QUERYSET in method_config:
                    assert callable(method_config[_KW_CONFIG_QUERYSET]), (
                        '`{}` field in endpoint `config` must be a callable accepting '
                        'parameters: `model` and `**url` in endpoint `{}`'
                    ).format(_KW_CONFIG_QUERYSET, clsname)

                    if _KW_CONFIG_FILTERS in method_config:
                        assert isinstance(method_config[_KW_CONFIG_FILTERS], dict), (
                            '`{}` field must be of type `dict`, containing pairs of `filter_name`, '
                            '`filter_func(queryset, value)` in endpoint `{}`'
                        ).format(_KW_CONFIG_FILTERS, clsname)

                elif method_name == 'delete' and not skip_model_checks:
                    raise AssertionError(('`config` for `delete` must include '
                                          '`{}` field for endpoint `{}`').format(_KW_CONFIG_QUERYSET, clsname))

                elif _KW_CONFIG_FILTERS in method_config:
                    raise AssertionError('`{}` cannot be used without `{}` in endpoint `{}`'
                                         .format(_KW_CONFIG_FILTERS, _KW_CONFIG_QUERYSET, clsname))

                if _KW_CONFIG_URL_FIELDS in method_config:
                    assert hasattr(method_config[_KW_CONFIG_URL_FIELDS], '__iter__') \
                           and not isinstance(method_config[_KW_CONFIG_URL_FIELDS], str), (
                        '`{}` config field must be an iterable in endpoint `{}`'
                    ).format(_KW_CONFIG_URL_FIELDS, clsname)

                if _KW_CONFIG_DO_NOT_NORMALIZE_DATA in method_config:
                    if _KW_CONFIG_DO_NOT_NORMALIZE_DATA is not True:
                        del method_config[_KW_CONFIG_DO_NOT_NORMALIZE_DATA]
        return config


class Endpoint(metaclass=MetaEndpoint):
    """
    Base class with default view and permission implementations
    """
    config = ddict()
    model, serializer = None, None

    def __call__(self, *args, **kwargs):
        """
        Produces a more descriptive error message, when a call
        to `.as_view()` is missed in the URL configuration.
        :raises ImproperlyConfigured
        """
        raise ImproperlyConfigured(('You are trying to call an `Endpoint` object directly.\n'
                                    'Did you forget to use `.as_view()` in the URL configuration?\n'
                                    'Offending endpoint: `{0}`'
                                    ).format(self.__class__.__qualname__))

    @classmethod
    def as_view(cls):
        """
        Acts as a main entry point for a request-response process,
        allowing to use `MyEndpoint.as_view()` directly in urlpatterns.
        :raises ImproperlyConfigured
        """
        if not len(cls.config):
            raise ImproperlyConfigured(('You are trying to call `as_view` on an endpoint with an empty config.\n'
                                        'Did you forget to specify the `config` attribute?\n'
                                        'Offending endpoint: `{0}`'
                                        ).format(cls.__qualname__))
        view_func = partial(ViewPrototype.respond, cls())
        view_func.__name__, view_func.__module__, view_func.__doc__ = cls.__name__, cls.__module__, cls.__doc__
        return api_view(tuple(cls.config.keys()))(view_func)

    @classmethod
    def make_serializer(cls, context: RequestContext, data=empty, many=None, **kwargs) -> ValidatedSerializer:
        """
        a DRY shortcut for quickly instantiating an assigned serializer.
        :param context: a `RequestContext` object to extract the queryset from
        :param data: request data that shall be serialized
        :param many: a bool flag, that when set, means more than one object will be serialized
        :return: serializer instance
        """
        extra_kwargs = dict(context=context) if issubclass(cls.serializer, ValidatedSerializer) else {}
        extra_kwargs.update(kwargs)
        if context.queryset is None:
            return cls.serializer(data=data,
                                  many=context.data_has_many if many is None else many,
                                  **extra_kwargs)
        return cls.serializer(instance=context.queryset,
                              data=data,
                              many=context.queryset_has_many if many is None else many,
                              **extra_kwargs)

    def _create_and_serialize(self, context, allow_many=undefined):
        serializer = self.make_serializer(context, context.data, context.data_has_many if allow_many else False)
        return ValidatedSerializer.validate_and_save(serializer)

    """
    Method independent logic
    """

    def transform_input_data(self, data: Union[ddict, list]) -> Union[ddict, list, None]:
        return data

    def transform_output_data(self, data: Union[ddict, list]) -> Union[ddict, list, None]:
        return data

    """
    Default view handler implementations
    """

    def on_get(self, context: RequestContext) -> Union[Response, int, dict, Tuple[dict, int]]:
        return self.make_serializer(context).data

    def on_post(self, context: RequestContext) -> Union[Response, int, dict, Tuple[dict, int]]:
        return self._create_and_serialize(context, self.config.post.allow_many), 201

    def on_put(self, context: RequestContext) -> Union[Response, int, dict, Tuple[dict, int]]:
        if context.queryset is None:
            return self._create_and_serialize(context, self.config.put.allow_many), 201

        serializer = self.make_serializer(context, context.data)
        return ValidatedSerializer.validate_and_save(serializer), 200

    def on_patch(self, context: RequestContext) -> Union[Response, int, dict, Tuple[dict, int]]:
        if context.queryset is None:
            raise EndpointError(status.HTTP_404_NOT_FOUND)

        serializer = self.make_serializer(context, context.data, partial=True)
        return ValidatedSerializer.validate_and_save(serializer), 200

    def on_delete(self, context: RequestContext) -> Union[Response, int, dict, Tuple[dict, int]]:
        serializer = self.make_serializer(context, context.data)
        serializer.delete()
        try:
            return serializer.data
        except AssertionError:
            return serializer.initial_data



from django_alt.utils.shortcuts import queryset_has_many
from django_alt.dotdict import ddict


class RequestContext:
    """
    Encapsulates data that is necessary to handle an `Endpoint` request
    """

    def __init__(self, request, *, data, queryset, query_params=None, url_args=None, url_kwargs=None):
        self.request = request
        self.queryset = queryset
        self.url_args = url_args
        self.url_kwargs = url_kwargs
        self.query_params = query_params
        self._data = data

    @property
    def queryset_has_many(self):
        return queryset_has_many(self.queryset)

    @property
    def data_has_many(self):
        return isinstance(self.data, list)

    @property
    def data(self):
        """
        Shorthand returning data property of the associated request
        """
        return self._data

    @property
    def user(self):
        return self.request.user


from rest_framework import serializers

from django_alt.managers import ValidatedManager


class ValidatedListSerializer(serializers.ListSerializer):
    """
    Overridden default `ListSerializer` class to support multiple delete.
    """
    def delete(self):
        assert self.instance is not None, (
            'Instances were not passed to `{}` constructor\n'
            'but `delete` was called.'
            'Did you forget to pass the queryset when constructing the serializer?'
        ).format(self.__class__.__qualname__)

        result = list(self.instance)
        for instance in result:
            self.child._do_delete_pre(instance)

        self.instance.delete()
        self.instance = None

        def remove_pk(i):
            i.pk = None
            return i

        self._validated_data = [
            self.child._do_delete_post(remove_pk(instance)) for instance in result
        ]
        return self.validated_data


class ValidatedSerializer(serializers.Serializer, ValidatedManager):
    def __init__(self, *args, model_class=None, validator_class=None, context=None, **kwargs):
        self.no_validation = kwargs.pop('no_validation', False)
        if self.no_validation:
            super().__init__(*args, **kwargs)
        else:
            if self.Meta:
                if model_class is None:
                    assert hasattr(self.Meta, 'model'), (
                        'Missing `model` field in serializer Meta class. '
                        'Offending serializer: {}').format(self.__class__.__qualname__)
                    model_class = self.Meta.model
                if validator_class is None:
                    assert hasattr(self.Meta, 'validator'), (
                        'Missing `validator` field in serializer Meta class. '
                        'Offending serializer: {}').format(self.__class__.__qualname__)
                    validator_class = self.Meta.validator

            super().__init__(*args, **kwargs)
            ValidatedManager.__init__(self, model_class=model_class, validator_class=validator_class, context=context, **kwargs)

    @classmethod
    def many_init(cls, *args, **kwargs):
        list_serializer = getattr(cls.Meta, 'list_serializer_class', None)
        if list_serializer is None:
            setattr(cls.Meta, 'list_serializer_class', ValidatedSerializer.Meta.list_serializer_class)
        return super().many_init(*args, **kwargs)

    @staticmethod
    def validate_and_save(serializer: serializers.Serializer, **kwargs):
        """
        a DRY Shorthand for executing `Serializer` subclass:
        - validation procedures
        - saving the model
        - returning serialized data
        :param serializer: rest_framework.serializers.Serializer subclass
        :param kwargs: extra kwargs to pass to serializer's `save` function
        :return: serializer.data
        """
        serializer.is_valid(raise_exception=True)
        serializer.save(**kwargs)
        return serializer.data

    def validate(self, attrs):
        self.make_validator(attrs, is_create=self.instance is None)
        return self.validate_only()

    def create(self, validated_data):
        return self.do_create(**validated_data)

    def update(self, instance, validated_data):
        return self.do_update(instance, **validated_data)

    def delete(self):
        return self.do_delete(self.instance, **self.initial_data)

    def to_representation(self, instance):
        if self.no_validation:
            return super().to_representation(instance)
        attrs = super().to_representation(instance)
        if self.validator is None:
            self.make_validator(attrs)
        else:
            self.validator.attrs = attrs
        self.validator.prepare_read_fields(instance)
        self.validator.will_read(instance)
        return self.validator.attrs

    class Meta:
        list_serializer_class = ValidatedListSerializer


class ValidatedModelSerializer(ValidatedSerializer, serializers.ModelSerializer):
    """
    Combines `ValidatedManager` and `rest_framework.serializers.ModelSerializer`.
    This construction allows overriding `create` and `update` functions without
    having to manually include the validator statements (`base`, `will_create`, etc.).
    Sequential validator calls are now handled by the `ValidatedManager`.
    Override `create` and `update` methods like this:
    
    >>> def create(self, validated_data):
    >>>     # for example you want to create a new record only under a certain condition
    >>>     if should_create_new():
    >>>         # actual creation is done by `do_create`
    >>>         return self.do_create()
    >>>     else:
    >>>         # skip creation and return an existing object
    >>>         return MyModel.objects.first()
    """



from django.test import TestCase
from django_alt.dotdict import ddict

from django_alt.validators import Validator


class ValidatorsMiscTests(TestCase):
    def test_attrs_setter_converts_to_ddict_positive(self):
        validator = Validator(dict(a=1))
        self.assertTrue(isinstance(validator.attrs, ddict))

        validator.attrs = dict(a=2)
        self.assertTrue(isinstance(validator.attrs, ddict))

    def test_is_create_accessed_when_it_was_not_passed_as_context_negative(self):
        validator = Validator(dict(a=1))
        with self.assertRaises(AssertionError) as ex:
            validator.is_create
        self.assertIn('`is_create` or `is_update` accessor was invoked. Offending validator: `Validator`',
                      ex.exception.args[0])

    def test_is_create_accessed_positive(self):
        validator = Validator(dict(a=1), is_create=True)
        self.assertTrue(validator.is_create)

    def test_is_update_accessed_when_it_was_not_passed_as_context_negative(self):
        validator = Validator(dict(a=1))
        with self.assertRaises(AssertionError) as ex:
            validator.is_update
        self.assertIn('`is_create` or `is_update` accessor was invoked. Offending validator: `Validator`',
                      ex.exception.args[0])

    def test_is_update_accessed_positive(self):
        validator = Validator(dict(a=1), is_create=True)
        self.assertFalse(validator.is_update)



from unittest import mock
from unittest.mock import PropertyMock

from django.test import TestCase
from django_alt.managers import ValidatedManager

from django_alt.validators import Validator
from django_alt_tests.conf.models import ModelA


class SampleValidator(Validator):
    def __init__(self, attrs, *, model=None, **context):
        self.calls = []
        super().__init__(attrs=attrs, model=model, **context)

    def default_field_1(self):
        return 'default'

    def clean_field_1(self, v):
        self.calls.append('clean_field_1')
        return v

    def clean_field_2(self, v):
        self.calls.append('clean_field_2')
        return v

    def field_field_1(self, v):
        self.calls.append('field_field_1')

    def field_field_2(self, v):
        self.calls.append('field_field_2')

    def check_x(self):
        self.calls.append('check_x')


class SampleManager(ValidatedManager):
    pass


class ValidatedManagerTestCase(TestCase):
    def assert_called(self, mock, what):
        return next(fun for fun in mock.mock_calls if what in str(fun))

    @mock.patch('tests.test_managers.SampleValidator')
    def test_create(self, mock):
        manager = SampleManager(ModelA, mock)
        attrs = dict(field_2=42, field_1='abc')
        type(mock.return_value).attrs = PropertyMock(return_value=attrs)

        instance = manager.do_create(**attrs)

        self.assert_called(mock, 'pre')
        self.assert_called(mock, 'clean')
        self.assert_called(mock, 'base')
        self.assert_called(mock, 'base_db')
        self.assert_called(mock, 'validate_fields')
        self.assert_called(mock, 'validate_checks')
        self.assert_called(mock, 'clean_and_default_fields')
        will_create_ = self.assert_called(mock, 'will_create')
        self.assert_called(mock, 'will_create_or_update')
        did_create_ = self.assert_called(mock, 'did_create')
        did_create_or_update = self.assert_called(mock, 'did_create_or_update')
        self.assert_called(mock, 'post')

        self.assertEqual(will_create_[1], ())
        self.assertEqual(did_create_[1], (instance,))
        self.assertEqual(did_create_or_update[1], (instance,))

        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().field_2, 42)
        self.assertEqual(ModelA.objects.first().field_1, 'abc')

    def test_create_deep_inspection(self):
        manager = SampleManager(ModelA, SampleValidator)
        attrs = dict(field_2=42, field_1='abc')
        manager.do_create(**attrs)
        assert len(set(manager.validator.calls).difference({'clean_field_1',
                                                            'clean_field_2',
                                                            'field_field_1',
                                                            'field_field_2',
                                                            'check_x'})) == 0

    @mock.patch('tests.test_managers.SampleValidator')
    def test_update(self, mock):
        instance = ModelA.objects.create(field_2=42, field_1='ltu')
        manager = SampleManager(ModelA, mock)
        attrs = dict(field_2=1337, field_1='ukr')
        type(mock.return_value).attrs = PropertyMock(return_value=attrs)

        instance = manager.do_update(instance, **attrs)

        self.assert_called(mock, 'pre')
        self.assert_called(mock, 'clean')
        self.assert_called(mock, 'base')
        self.assert_called(mock, 'base_db')
        self.assert_called(mock, 'validate_fields')
        self.assert_called(mock, 'validate_checks')
        self.assert_called(mock, 'clean_and_default_fields')
        will_update_ = self.assert_called(mock, 'will_update')
        self.assert_called(mock, 'will_create_or_update')
        did_update_ = self.assert_called(mock, 'did_update')
        did_create_or_update = self.assert_called(mock, 'did_create_or_update')
        self.assert_called(mock, 'post')

        self.assertEqual(will_update_[1], (instance,))
        self.assertEqual(did_update_[1], (instance,))
        self.assertEqual(did_create_or_update[1], (instance,))

        will_update_ = next(fun for fun in mock.mock_calls if 'will_update' in str(fun))
        did_update_ = next(fun for fun in mock.mock_calls if 'did_update' in str(fun))

        self.assertEqual(will_update_[1], (instance,))
        self.assertEqual(did_update_[1], (instance,))

        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().field_2, 1337)
        self.assertEqual(ModelA.objects.first().field_1, 'ukr')

    def test_update_deep_inspection(self):
        instance = ModelA.objects.create(field_2=42, field_1='ltu')
        manager = SampleManager(ModelA, SampleValidator)
        attrs = dict(field_2=1337, field_1='ukr')
        instance = manager.do_update(instance, **attrs)
        assert len(set(manager.validator.calls).difference({'clean_field_1',
                                                            'clean_field_2',
                                                            'field_field_1',
                                                            'field_field_2',
                                                            'check_x'})) == 0

    @mock.patch('tests.test_managers.SampleValidator')
    def test_delete(self, mock):
        instance = ModelA.objects.create(field_2=42, field_1='ltu')
        manager = SampleManager(ModelA, mock)
        self.assertEqual(ModelA.objects.count(), 1)

        attrs = dict(arbitrary='field')
        type(mock.return_value).attrs = PropertyMock(return_value=attrs)

        instance = manager.do_delete(instance, **attrs)

        self.assert_called(mock, 'pre')
        will_delete_ = next(fun for fun in mock.mock_calls if 'will_delete' in str(fun))
        did_delete_ = next(fun for fun in mock.mock_calls if 'did_delete' in str(fun))
        self.assert_called(mock, 'post')

        self.assertEqual(will_delete_[1], (instance,))
        self.assertEqual(did_delete_[1], (instance,))
        self.assertEqual(ModelA.objects.count(), 0)

    def test_calling_validate_only_without_a_validator_negative(self):
        manager = SampleManager(ModelA, SampleValidator)
        with self.assertRaises(AssertionError) as ex:
            manager.validate_only()

        self.assertIn('call `make_validator` on `SampleManager` before', ex.exception.args[0])

    def test_create_sets_is_create_flag_on_validator_positive(self):
        manager = SampleManager(ModelA, SampleValidator)
        manager.do_create(field_1='a', field_2=1)
        self.assertTrue(manager.validator.is_create)
        self.assertFalse(manager.validator.is_update)

    def test_create_sets_is_update_flag_on_validator_positive(self):
        manager = SampleManager(ModelA, SampleValidator)
        manager.do_create(field_1='a', field_2=1)
        self.assertTrue(manager.validator.is_create)
        self.assertFalse(manager.validator.is_update)
        manager.do_update(ModelA.objects.first(), field_1='a', field_2=1)
        self.assertTrue(manager.validator.is_update)
        self.assertFalse(manager.validator.is_create)

    def test_create_sets_default_field_positive(self):
        manager = SampleManager(ModelA, SampleValidator)
        obj = manager.do_create(field_2=1)
        self.assertEqual(obj.field_1, 'default')



import collections

from django.db.models import QuerySet
from rest_framework import serializers

validation_error_class = serializers.ValidationError


def invalid_if(condition, key_or_list, error_or_list):
    """
    Shortcut for raising a validation error if a condition is met.
    :raises: serializers.ValidationError
    """
    if condition:
        invalid(key_or_list, error_or_list)


def invalid(key_or_list, error_or_list):
    """
    Shortcut for raising a validation error.
    :raises: serializers.ValidationError
    """
    raise validation_error_class(make_error(key_or_list, error_or_list))


def is_iterable(obj):
    """
    Shortcut for checking if object is an iterable.
    This method *does not* count strings as iterables.
    :param obj: object to check
    :return: whether the object is an iterable or not
    """
    return not isinstance(obj, str) and isinstance(obj, collections.Iterable)


def make_error(key_or_list, error_or_list) -> dict:
    """
    Creates a unified error object.
    :param key_or_list: field name (can be None if error is a non_field one or a list of str)
    :param error_or_list: one or more string representations of the error
    """
    err = error_or_list if is_iterable(error_or_list) else [error_or_list]
    err = list(map(lambda e: (e if e[-1] == '.' or e[-1] == '!' else e + '.') if len(e) else e, err))
    if key_or_list:
        return {k: err for k in key_or_list} if is_iterable(key_or_list) else {key_or_list: err}
    return {'non_field_errors': err}


def if_in(key, container, func_true=None, default=None):
    """
    Checks if given key on the container exists and executes.
    func_true or func_false with each corresponding value based on the outcome.
    :param key: the key to search
    :param container: the container to search in
    :param func_true: callable that is called on success: func_true(container[key])
    :param default: default value to set if key does not exist in the container
    :return: {bool} whether the key exists
    """
    if key in container and func_true is not None:
        container[key] = func_true(container[key])
        return True
    elif default is not None:
        container[key] = default
        return False


def queryset_has_many(queryset) -> bool:
    """
    Checks whether a given query product is iterable.
    :param queryset: Django QuerySet object
    :return: {bool}
    """
    return isinstance(queryset, QuerySet) or hasattr(queryset, '__iter__')


def coal(obj, fallback):
    """
    Inspired by C# null-coalescing operator:
    C# --> not_null = obj ?? new Object()
    Py --> not_null = coal(obj, Object())
    Also returns fallback upon a KeyError
    """
    try:
        return obj if obj is not None else fallback
    except KeyError:
        return fallback


def coald(obj, fallback_lambda):
    """
    A deferred variation of the null-coalescing operator.
    Returns obj if it is defined, execution result of a parametress lambda (fallback_lambda) otherwise.
    """
    return obj if obj is not None else fallback_lambda()


def if_all_in(keys, container, func_true=None):
    """
    checks if all given keys are in the container and
    executes func_true for each of the corresponding values
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether the keys exist
    """
    if all(key in container for key in keys):
        for key in keys:
            container[key] = func_true(container[key])
        return True
    return False


def if_any_in(keys, container, func_true=None):
    """
    checks if any given keys ar in the container and
    executes func_true for each of the corresponding values of the existing keys
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether any of the keys exist
    """
    found_key = False
    for key in keys:
        if key in container:
            container[key] = func_true(container[key])
            found_key = True
    return found_key


def first_defined(*args):
    """
    Finds the first argument that is not None
    :param args: arguments to check for "defineness"
    :return:
    """
    for arg in args:
        if arg is not None:
            return arg
    return None


def prohibited(key: str, container: dict = None):
    """
    shortcut for raising a validation error when a prohibited field is in the container.
    :param container: container to search
    :param key: existing key
    :raises: serializer.ValidationError
    """
    invalid_if(container and key in container, key, 'This field cannot be present.')


def prohibited_any(keys: tuple, container: dict = None):
    """
    shortcut for raising a validation error when any of the given fields exist in the container.
    :param container: container to search
    :param key: existing key
    :raises: serializer.ValidationError
    """
    intersection = set(keys).intersection(container.keys())
    if container and len(intersection) > 0:
        invalid(intersection, 'This field cannot be present.')


def required(key: str, container: dict = None):
    """
    shortcut for raising a validation error about a missing required field.
    :param container: container to search
    :param key: missing key
    :raises: serializer.ValidationError
    """
    if container and key in container:
        return container[key]
    invalid(key, 'This field is required.')


def required_all(keys: list, container: dict = None):
    """
    shortcut for raising a validation error if any of the keys are missing in the container.
    :param keys: keys that are required in the container
    :param container: container to search
    :raises: serializer.ValidationError
    """
    if container:
        difference = set(keys).difference(container.keys())
        if len(difference):
            invalid(difference, 'This field is required.')


def try_cast(typ, value):
    """
    Attempts to cast value to a given type.
    :param typ: type to cast to
    :param value: value to be casted
    :return: casted value if cast was successful, otherwise None
    """
    try:
        return typ(value)
    except (ValueError, TypeError):
        return None


def valid_if(condition, key_or_list, error_or_list):
    """
    Shortcut for raising a validation error if a condition is not met.
    Think about this function as an assertion (i.e. this condition must be met).
    :raises: serializers.ValidationError
    """
    if not condition:
        invalid(key_or_list, error_or_list)


"""
Alias of `valid_if`.
Used in Design by Contract methodology to denote
that a callable asserts that a given condition is true
before executing any code
"""
expects = valid_if

"""
Alias of `valid_if`.
Used in Design by Contract methodology to denote
that a callable asserts that a given condition is true
after its code block is executed
"""
ensures = valid_if



from .abstract.validators import Validator
from .utils.shortcuts import coal


class ValidatedManager:
    """
    Relates validator to ObjectManager, allowing to easily use validator
    without needing a serializer
    """

    def __init__(self, model, validator_class: Validator, no_save=False, **context):
        self.model = model
        self.no_save = no_save
        self.validator = validator_class(model=model, serializer=None, **context)

    def validation_sequence(self, attrs: dict):
        self.validator.clean_fields(attrs, attrs.keys())

        attrs = coal(self.validator.clean(attrs), attrs)
        attrs = coal(self.validator.base(attrs), attrs)

        self.validator.validate_fields(attrs, attrs.keys())
        self.validator.validate_checks(attrs)

    def create(self, **attrs):
        """
        Validates and creates a model instance
        :param attrs: attributes to create the instance from.
        :return: the newly created instance
        """
        self.validation_sequence(attrs)

        attrs = coal(self.validator.will_create(attrs), attrs)
        attrs = coal(self.validator.base_db(attrs), attrs)

        if not self.no_save:
            instance = self.model.objects.create(**attrs)
            self.validator.did_create(instance, attrs)
            return instance

        return attrs

    def update(self, instance, **attrs):
        """
        Validates and creates a model instance
        :param attrs: attributes to create the instance from.
        :return: the newly created instance
        """
        self.validation_sequence(attrs)

        attrs = coal(self.validator.will_update(instance, attrs), attrs)
        attrs = coal(self.validator.base_db(attrs), attrs)

        if not self.no_save:
            for k, v in attrs.items():
                setattr(instance, k, v)
            instance.save()
            self.validator.did_update(instance, attrs)
            return instance

        return attrs

    def delete(self, queryset):
        self.validator.will_delete(queryset)
        queryset.delete()
        self.validator.did_delete()

    def create_many(self, list_of_attrs):
        """
        TBA
        :param list_of_attrs:
        :return:
        """
        instances = []
        list_of_attrs = list(list_of_attrs)
        for attrs in list_of_attrs:
            self.validation_sequence(attrs)
            instances.append(self.model(**attrs))

        self.model.objects.bulk_create(instances)
        for instance, attrs in zip(instances, list_of_attrs):
            self.validator.did_create(instance, attrs)

        return instances



import collections
from typing import Union

from django.db.models import QuerySet
from rest_framework import serializers

ValidationError = serializers.ValidationError


def as_bool(string: str) -> Union[bool, None]:
    """
    Shortcut for checking if string is actually a representation for bool.
    This is useful when i.e. passing bool values by query string.
    :param string: input to check
    :return: bool value if string matches, None otherwise
    """
    lower = string.lower()
    if lower == 'true':
        return True
    elif lower == 'false':
        return False
    return None


def copyattrs(source, destination, *attrs):
    """
    Shortcut for getting multiple attributes from a source object and
    setting them on the destination object.
    Useful for when you want to set multiple ddict attributes on a class instance.
    :param source: object that will have getattr called on
    :param destination: object that will have setattr called on
    :param attrs: list of attributes to iterate on
    :return: destination object
    """
    for attr in attrs:
        setattr(destination, attr, getattr(source, attr))
    return destination


def invalid_if(condition, key_or_list, error_or_list):
    """
    Shortcut for raising a validation error if a condition is met.
    :raises: serializers.ValidationError
    """
    return not condition or invalid(key_or_list, error_or_list)


def invalid(key_or_list, error_or_list):
    """
    Shortcut for raising a validation error.
    :raises: serializers.ValidationError
    """
    raise ValidationError(make_error(key_or_list, error_or_list))


def is_iterable(obj):
    """
    Shortcut for checking if object is an iterable.
    This method *does not* count strings as iterables.
    :param obj: object to check
    :return: whether the object is an iterable or not
    """
    if not isinstance(obj, str):
        try:
            iter(obj)
            return True
        except TypeError:
            pass
    return False


def make_error(key_or_list, error_or_list) -> dict:
    """
    Creates a unified error object.
    :param key_or_list: field name (can be None if error is a non_field one or a list of str)
    :param error_or_list: one or more string representations of the error
    """
    err = error_or_list if is_iterable(error_or_list) else [error_or_list]
    err = list(map(lambda e: (e if e[-1] == '.' or e[-1] == '!' else e + '.') if len(e) else e, err))
    if key_or_list:
        return {k: err for k in key_or_list} if is_iterable(key_or_list) else {key_or_list: err}
    return {'non_field_errors': err}


def if_in(key, container, func_true=None, default=None):
    """
    Checks if given key on the container exists and executes.
    func_true or func_false with each corresponding value based on the outcome.
    :param key: the key to search
    :param container: the container to search in
    :param func_true: callable that is called on success: func_true(container[key])
    :param default: default value to set if key does not exist in the container
    :return: {bool} whether the key exists
    """
    if key in container and func_true is not None:
        container[key] = func_true(container[key])
        return True
    elif default is not None:
        container[key] = default
        return False


def queryset_has_many(queryset) -> bool:
    """
    Checks whether a given query product is iterable.
    :param queryset: Django QuerySet object
    :return: {bool}
    """
    return isinstance(queryset, QuerySet) or is_iterable(queryset)


def coal(obj, fallback):
    """
    Inspired by C# null-coalescing operator:
    C# --> not_null = obj ?? new Object()
    Py --> not_null = coal(obj, Object())
    Also returns fallback upon a KeyError
    """
    try:
        return obj if obj is not None else fallback
    except KeyError:
        return fallback


def coald(obj, fallback_lambda):
    """
    A deferred variation of the null-coalescing operator.
    Returns obj if it is defined, execution result of a parametress lambda (fallback_lambda) otherwise.
    """
    return obj if obj is not None else fallback_lambda()


def if_all_in(keys, container, func_true=None):
    """
    checks if all given keys are in the container and
    executes func_true for each of the corresponding values
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether the keys exist
    """
    if all(key in container for key in keys):
        for key in keys:
            container[key] = func_true(container[key])
        return True
    return False


def if_any_in(keys, container, func_true=None):
    """
    checks if any given keys ar in the container and
    executes func_true for each of the corresponding values of the existing keys
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether any of the keys exist
    """
    found_key = False
    for key in keys:
        if key in container:
            container[key] = func_true(container[key])
            found_key = True
    return found_key


def first_defined(*args):
    """
    Finds the first argument that is not None
    :param args: arguments to check for "definedness"
    :return:
    """
    for arg in args:
        if arg is not None:
            return arg
    return None


def pluck(*keys, container: dict):
    """
    Creates a new dict from a given one by selecting only desired keys.
    :param container: container to pluck
    :param keys: keys to take
    :return: new dictionary
    """
    return {k: v for (k, v) in container.items() if k in keys}


def prohibited(key: str, container: dict = None):
    """
    shortcut for raising a validation error when a prohibited field is in the container.
    :param container: container to search
    :param key: existing key
    :raises: serializer.ValidationError
    """
    invalid_if(container and key in container, key, 'This field cannot be present.')


def prohibited_any(keys: tuple, container: dict = None):
    """
    shortcut for raising a validation error when any of the given fields exist in the container.
    :param keys: keys that cannot be in the container
    :param container: container to search
    :raises: serializer.ValidationError
    """
    if container:
        intersection = set(keys).intersection(container.keys())
        if len(intersection) > 0:
            invalid(intersection, 'This field cannot be present.')


def required(key: str, container: dict = None):
    """
    shortcut for raising a validation error about a missing required field.
    :param key: missing key
    :param container: container to search
    :raises: serializer.ValidationError
    """
    if container and key in container:
        return container[key]
    invalid(key, 'This field is required.')


def required_all(keys: list, container: dict = None):
    """
    shortcut for raising a validation error if any of the keys are missing in the container.
    :param keys: keys that are required in the container
    :param container: container to search
    :raises: serializer.ValidationError
    """
    if container:
        difference = set(keys).difference(container.keys())
        if len(difference):
            invalid(difference, 'This field is required.')


def try_cast(typ, value):
    """
    Attempts to cast value to a given type.
    :param typ: type to cast to
    :param value: value to be casted
    :return: casted value if cast was successful, otherwise None
    """
    try:
        return typ(value)
    except (ValueError, TypeError):
        return None


def valid_if(condition, key_or_list, error_or_list):
    """
    Shortcut for raising a validation error if a condition is not met.
    Think about this function as an assertion (i.e. this condition must be met).
    :raises: serializers.ValidationError
    :returns: the condition variable, if it is truthy
    """
    return condition or invalid(key_or_list, error_or_list)


"""
Alias of `valid_if`.
Used in Design by Contract methodology to denote
that a callable asserts that a given condition is true
before executing any code
"""
expects = valid_if

"""
Alias of `valid_if`.
Used in Design by Contract methodology to denote
that a callable asserts that a given condition is true
after its code block is executed
"""
ensures = valid_if

"""
TODO:
dict validators as higher order functions, i.e.:

mydict = dict(a=1, b=2)
required = make_required(mydict)
required('a')

or maybe even better and more legible

dict_validator = make_validator(mydict)
dict_validator.required('a')
dict_validator.required_all('a', 'b', 'd')
dict_validator.if_in('a', lambda a: a.lower())

new shortcuts:
    def not_null(value)
    def is_lte(value: Union[str, int], boundary)
        ''' string length and number value '''
    def is_gte(value: Union[str, int], boundary)
    def is_lt
    def is_gt
"""



from itertools import chain


class _undefined_meta(type):
    def __bool__(self):
        return False


class undefined(metaclass=_undefined_meta):
    """
    An empty class that designates a non-existing ddict item.
    Used for graceful item retrieval.
    Inspired by JavaScript.
    """

    def __new__(cls, *args, **kwargs):
        if cls is undefined:
            raise TypeError('Class `undefined` cannot be instantiated')

    def __bool__(self):
        return False


class ddict(dict):
    def __init__(self, iterable=None, **kwargs):
        if iterable is not None:
            if isinstance(iterable, dict):
                iterable.update(kwargs)
            else:
                iterable = chain(iterable, kwargs.items())
            super().__init__(iterable)
        else:
            super().__init__(**kwargs)

    def __iter__(self):
        return self.items().__iter__()

    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            return undefined

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __add__(self, other):
        if isinstance(other, dict):
            result = self.copy()
            result.update(other)
            return result
        raise TypeError('You can only add two ddict or one ddict one dict instances')



from typing import Type

from django.db.models import Model
from django_alt.validators import Validator

from django_alt.contexts import RequestContext


class ValidatedManager:
    def __init__(self, model_class: Type[Model],
                 validator_class: Type[Validator], context: Type[RequestContext] = None, **kwargs):
        self.request_context = context
        self.model_class = model_class
        self.validator_class = validator_class
        self.validator = None

    def make_validator(self, attrs, **kwargs):
        kwargs.setdefault('model', self.model_class)
        kwargs.setdefault('context', self.request_context)
        self.validator = self.validator_class(attrs, **kwargs)
        return self.validator

    def validate_only(self):
        assert self.validator is not None, (
            'You have to call `make_validator` on `{}` before explicitly calling `validate_only`.'
        ).format(self.__class__.__name__)
        self.validator.pre()

        self.validator.clean_and_default_fields()
        self.validator.clean()

        self.validator.base()
        self.validator.validate_fields()
        self.validator.validate_checks()
        return self.validator.attrs

    def do_create(self, **attrs):
        """
        Validates given attribute dict and creates an object from it.
        :param attrs: attributes to create the object from
        :return: created instance
        """
        if self.validator is None:
            self.make_validator(attrs, is_create=True)
            self.validate_only()
        elif len(attrs):
            self.validator._is_create = True
            self.validator.attrs = attrs

        self.validator.will_create()
        self.validator.will_create_or_update()
        self.validator.base_db()

        instance = self.create_instance()

        self.validator.did_create(instance)
        self.validator.did_create_or_update(instance)
        self.validator.post()

        return instance

    def create_instance(self):
        return self.model_class.objects.create(**self.validator.attrs) if self.model_class is not None else None

    def do_update(self, instance, **attrs):
        """
        Validates given attribute dict and replaces fields of
        an existing object with its contents.
        :param instance: instance to update
        :param attrs: attributes to update the object with
        :return: updated instance
        """
        if self.validator is None:
            self.make_validator(attrs, is_create=False)
            self.validate_only()
        elif len(attrs):
            self.validator._is_create = False
            self.validator.attrs = attrs

        self.validator.will_update(instance)
        self.validator.will_create_or_update()
        self.validator.base_db()

        instance = self.update_instance(instance)

        self.validator.did_update(instance)
        self.validator.did_create_or_update(instance)
        self.validator.post()

        return instance

    def update_instance(self, instance):
        for k, v in self.validator.attrs.items():
            setattr(instance, k, v)
        instance.save()
        return instance

    def _do_delete_pre(self, queryset, **attrs):
        if self.validator is None:
            self.make_validator(attrs)
            self.validator.pre()

        self.validator.will_delete(queryset)

    def _do_delete_post(self, queryset, **attrs):
        self.validator.did_delete(queryset)
        self.validator.post()
        return queryset

    def delete_instance(self, instance):
        assert instance is not None, (
            'An instance was not passed to `{}` for deletion\n'
            'but `delete` was called.'
            'Did you forget to include a `query` clause in an endpoint `delete` config?'
        ).format(self.__class__.__qualname__)
        return instance.delete()

    def do_delete(self, queryset, **attrs):
        """
        Validates given attribute dict and calls delete on given queryset.
        :param queryset: object instance(s) to delete
        :param attrs: extra attributes to pass to the validation engine
        :return: deleted instance(s)
        """
        self._do_delete_pre(queryset, **attrs)
        queryset = self.delete_instance(queryset)
        self._do_delete_post(queryset, **attrs)
        return queryset



from django.test import TestCase

from django_alt.dotdict import ddict, undefined


class DdictTests(TestCase):
    def test_construction_from_dict(self):
        d = ddict({'a': 1, 'b': 2})
        self.assertEqual(len(d), 2)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)

    def test_construction_from_ddict(self):
        d = ddict(ddict({'a': 1, 'b': 2}))
        self.assertEqual(len(d), 2)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)

    def test_construction_from_iterable(self):
        d = ddict({'a': 1, 'b': 2}.items())
        self.assertEqual(len(d), 2)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)

    def test_construction_from_kwargs(self):
        d = ddict(a=1, b=2)
        self.assertEqual(len(d), 2)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)

    def test_construction_from_dict_and_kwargs(self):
        d = ddict({'c': 3}, a=1, b=2)
        self.assertEqual(len(d), 3)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d.c, 3)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)
        self.assertEqual(d['c'], 3)

    def test_construction_from_iterable_and_kwargs(self):
        d = ddict((('c', 3),), a=1, b=2)
        self.assertEqual(len(d), 3)
        self.assertEqual(d.a, 1)
        self.assertEqual(d.b, 2)
        self.assertEqual(d.c, 3)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)
        self.assertEqual(d['c'], 3)

    def test_del(self):
        d = ddict(a=1, b=2)
        self.assertEqual(len(d), 2)
        del d.a
        self.assertEqual(len(d), 1)
        del d['b']
        self.assertEqual(len(d), 0)

    def test_nonexistent_key(self):
        d = ddict(a=1)
        self.assertTrue(d.b is undefined)

    def test_iter(self):
        result = {}
        for k, v in ddict(a=1, b=2):
            result[k] = v

        self.assertDictEqual(result, {'a': 1, 'b': 2})

        result = {}
        for k, v in ddict(a=1, b=2).items():
            result[k] = v

        self.assertDictEqual(result, {'a': 1, 'b': 2})

    def test_equality(self):
        a, b = ddict(a=1, b=2), dict(a=1, b=2)
        self.assertTrue(a == b)
        self.assertEqual(a, b)
        self.assertEqual(a, b)

        a, b = ddict(b=2, a=1), dict(a=1, b=2)
        self.assertTrue(a == b)
        self.assertEqual(a, b)
        self.assertEqual(a, b)

        a, b = ddict(b=2, a=1, c=3), dict(a=1, b=2)
        self.assertTrue(a != b)
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, b)

        a, b = ddict(b=2, a=3), dict(a=1, b=2)
        self.assertTrue(a != b)
        self.assertNotEqual(a, b)
        self.assertNotEqual(a, b)

    def test_add_operator(self):
        a, b = ddict(b=2, a=1, c=3), dict(a=2, b=3)
        with self.assertRaises(TypeError):
            c = a + 1

        c = a + b
        self.assertDictEqual(c, dict(a=2, b=3, c=3))

        b2 = dict(z=4)
        c = a + b2
        self.assertDictEqual(a, dict(a=1, b=2, c=3))
        self.assertDictEqual(b2, dict(z=4))
        self.assertDictEqual(c, dict(a=1, b=2, c=3, z=4))


class UndefinedTests(TestCase):
    def test_undefined_is_undefined(self):
        self.assertTrue(undefined is undefined)
        self.assertFalse(undefined is False)
        self.assertFalse(undefined is True)
        self.assertFalse(undefined is 1)
        self.assertFalse(undefined is 0)
        a = undefined
        self.assertTrue(a is undefined)

    def test_cannot_construct_undefined(self):
        with self.assertRaises(TypeError):
            undefined()

    def test_undefined_casts_to_false(self):
        self.assertFalse(bool(undefined))
        self.assertFalse(undefined)



import inspect
from abc import abstractmethod
from collections import OrderedDict
from functools import partial

from typing import Type

from django.db.models import Model

from django_alt.dotdict import ddict, undefined
from django_alt.utils import shortcuts


class Phasers:
    """
    Provides a standard interface for encapsulating code called during a specific validation phase.
    Inspired by Perl 6 phasers.
    """

    @abstractmethod
    def pre(self):
        """
        Function that is called before a validation sequence begins.
        Use this for
        - placing preconditional validation logic (`required` statements, etc.)
        Still, you should avoid using heavy validation logic here.
        """
        pass

    @abstractmethod
    def post(self):
        """
        Function that is called last after a full (successful) validation sequence.
        Use this for
        - placing extra cleanup code
        - code that should be executed after a validation sequence regardless whether an object
          was created/read/updated or deleted.
        """
        pass


class LifecycleHooks:
    """
    Provides a standard interface for handling object lifecycle.
    """

    """
    Precursor lifecycle hooks.
    Invoked before a database operation.
    """

    @abstractmethod
    def will_create(self):
        """
        Called when a new instance is created from attrs.
        Use this for
        - placing validation logic that should be solely executed on record creation
        """
        pass

    @abstractmethod
    def will_read(self, instance: Type[Model]):
        pass

    @abstractmethod
    def will_update(self, instance: Type[Model]):
        """
        Called when an existing instance is populated with new attrs.
        Use this for
        - placing validation logic that should be solely executed on record update
        :param instance: existing instance that will be updated
        """
        pass

    @abstractmethod
    def will_delete(self, instance: Type[Model]):
        """
        Called before a queryset is deleted
        :param instance: items to be deleted
        """
        pass

    @abstractmethod
    def will_create_or_update(self):
        """
        Called before an instance is created from or updated with attrs.
        """
        pass

    """
    Postcursor lifecycle hooks.
    Invoked after a database operation.
    """

    @abstractmethod
    def did_create(self, instance: Type[Model]):
        """
        Called after a model instance is created
        :param instance: the created model instance
        """
        pass

    @abstractmethod
    def did_update(self, instance: Type[Model]):
        """
        Called after a model instance is updated
        :param instance: the updated model instance
        """
        pass

    @abstractmethod
    def did_delete(self, instance: Type[Model]):
        """
        Called after a model instance is deleted
        :param instance: queryset that was deleted
        """
        pass

    @abstractmethod
    def did_create_or_update(self, instance: Type[Model]):
        """
        Called after a model instance is created or an existing instance updated
        :param instance: the created or updated model instance
        """
        pass


class Validator(LifecycleHooks, Phasers):
    """
    Abstract class that defines the basic lifecycle hooks and definition
    principles for its subclasses
    """
    ATTR_CHECKS_PREFIX = 'check_'

    FIELD_CLEAN_PREFIX = 'clean_'
    FIELD_DEFAULT_PREFIX = 'default_'
    FIELD_VALIDATOR_PREFIX = 'field_'
    FIELD_READ_PREP_PREFIX = 'read_'

    def __init__(self, attrs, *, model=None, context=None, **kwargs):
        """
        :param [model]: model class of the serialized object (if serialized by a ModelSerializer)
        :param [context]: any data that gets passed as serializer kwargs
        """
        self._attrs = ddict(attrs)
        self.model = model
        self.context = context

        self._is_create = kwargs.get('is_create', None)

    @property
    def is_create(self) -> bool:
        """
        Returns whether the currently validated operation is a part of resource creation.
        :return: is creation validated
        """
        return self._is_create

    @property
    def is_update(self) -> bool:
        """
        Returns whether the currently validated operation is a part of existing resource update.
        :return: is update validated
        """
        return (not self.is_create) if self.is_create is not None else None

    @property
    def attrs(self) -> ddict:
        return self._attrs

    @attrs.setter
    def attrs(self, value):
        if not isinstance(value, OrderedDict):
            self._attrs = ddict(value)
        else:
            self._attrs = value

    def validate_checks(self):
        """
        If subclass defines functions with names starting with check_,
        executes such functions with attrs dict as the parameter.
        All functions are called in alphabetical order.
        """

        def is_check(n):
            return n.startswith(self.ATTR_CHECKS_PREFIX) and callable(getattr(self, n))

        for name in (name for name in dir(self) if is_check(name)):
            getattr(self, name)()

    def validate_fields(self):
        """
        If subclass defines functions named field_<field_name>
        where field_name corresponds to a field declared on the
        serializer, executes such functions with the field's value
        as the parameter.
        :return: None
        """
        for field in sorted(self.attrs.keys()):
            name = ''.join((self.FIELD_VALIDATOR_PREFIX, field))
            if hasattr(self, name) and callable(getattr(self, name)):
                getattr(self, name)(self.attrs[field])

    def clean_and_default_fields(self):
        """
        If subclass defines functions named clean_<field_name>
        where field_name corresponds to a attribute present in the attrs
        dict, executes such functions with the attribute's value
        as the parameter and sets the return value on the attrs dict
        :return: None
        """
        if self._is_create is not None and self._is_create is True:
            for default_func in (f for f in dir(self) if
                                 f.startswith(self.FIELD_DEFAULT_PREFIX) and callable(getattr(self, f))):
                field = default_func[len(self.FIELD_DEFAULT_PREFIX):]
                if field not in self.attrs:
                    self.attrs[field] = getattr(self, default_func)()

        for field in sorted(self.attrs.keys()):
            name = ''.join((self.FIELD_CLEAN_PREFIX, field))
            if hasattr(self, name) and callable(getattr(self, name)):
                self.attrs[field] = getattr(self, name)(self.attrs[field])

    def prepare_read_fields(self, instance):
        """
        If subclass defines functions named read_<field_name>
        where field_name corresponds to a field declared on the
        serializer, executes such functions with the field's
        representation value and read instance as the parameter.
        Such functions are expected to returned the prepared value
        that is written to the representation dict.
        :return: None
        """
        read_funcs = ((name, method) for name, method in inspect.getmembers(self, predicate=inspect.ismethod) \
                      if name.startswith(self.FIELD_READ_PREP_PREFIX))
        for field, func in read_funcs:
            self.attrs[field[len(self.FIELD_READ_PREP_PREFIX):]] = func(self.attrs.get(field, undefined), instance)

    @abstractmethod
    def clean(self):
        """
        Base clean method. Executed before base validation.
        Use this for
        - value cleaning (lowercasing, normalization, etc.)
        - setting or generating default dependent values (like slugs)
        :return: None
        """
        pass

    @abstractmethod
    def base(self):
        """
        Executed before create and update methods.
        Use this for
        - raising validation errors independent from create or update
        :return: None
        """
        pass

    @abstractmethod
    def base_db(self):
        """
        Executed as last step of validation. As DB access can be considered
        expensive, this will run last, provided any other steps check out.
        Use this for
        - validation that requires database access
        - validation logic that is more expensive
        :return: None
        """
        pass


class ContainerValidator(Validator):
    """
    Extends the default `Validator` by allowing an easier usage of dictionary validation
    shortcuts.
    """
    __empty__ = lambda *args, **kwargs: None
    required = required_all = pluck = __empty__

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        validators = (
            shortcuts.required, shortcuts.required_all, shortcuts.pluck
        )

        for validator in validators:
            setattr(self, validator.__name__, partial(validator, container=self.attrs))

    def rename_attr(self, key, new_key):
        """
        Pops the value at `key` in `self.attrs` and sets that value at `new_key`.
        Raises `KeyError` if `key` not found in `self.attrs`
        :param key: the key to find value at
        :param new_key: the key to set value on
        """
        self.attrs[new_key] = self.attrs.pop(key)



from unittest import TestCase

from django_alt.__legacy__.django_alt.utils.shortcuts import *


class UtilsShortcutsTests(TestCase):
    def test_make_error(self):
        self.assertEqual(make_error('k', 'v'), {'k': ['v.']})
        self.assertEqual(make_error('k', 'v.'), {'k': ['v.']})
        self.assertEqual(make_error(['k'], 'v'), {'k': ['v.']})
        self.assertEqual(make_error(['k'], ['v']), {'k': ['v.']})
        self.assertEqual(make_error(['k1', 'k2'], ['v']), {'k1': ['v.'], 'k2': ['v.']})
        self.assertEqual(make_error(['k1', 'k2'], ['v1', 'v2']), {'k1': ['v1.', 'v2.'], 'k2': ['v1.', 'v2.']})
        self.assertEqual(make_error(None, 'v1'), {'non_field_errors': ['v1.']})

    def test_invalid(self):
        with self.assertRaises(validation_error_class) as ex:
            invalid('k', 'v')
        self.assertEqual(ex.exception.detail, {'k': ['v.']})

    def test_invalid_if(self):
        invalid_if(False, 'k', 'v')
        with self.assertRaises(validation_error_class) as ex:
            invalid_if(True, 'k', 'v')
        self.assertEqual(ex.exception.detail, {'k': ['v.']})

    def test_coal(self):
        self.assertDictEqual(coal(None, {}), {})
        self.assertDictEqual(coal(None, {'a': 5}), {'a': 5})
        self.assertEqual(coal(1, {'a': 5}), 1)
        self.assertEqual(coal(False, {'a': 5}), False)
        self.assertEqual(coal(False, None), False)
        self.assertEqual(coal(None, False), False)
        self.assertEqual(coal(None, None), None)

    def test_if_in(self):
        a = dict(a=1, b=2)
        self.assertTrue(if_in('a', a, lambda a: a + 1))
        self.assertEqual(a['a'], 2)
        self.assertFalse(if_in('c', a, lambda c: c + 1, 15))
        self.assertEqual(a['c'], 15)

    def test_first_defined(self):
        self.assertEqual(first_defined(
            False
        ), False)

        self.assertEqual(first_defined(
            None,
            False
        ), False)

        self.assertEqual(first_defined(
            None,
            False,
            'foo',
        ), False)

    def test_if_all_in(self):
        a = dict(a=1, b=2, c=3)
        self.assertTrue(if_all_in(['a', 'b', 'c'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=2, b=3, c=4))

        self.assertFalse(if_all_in(['a', 'b', 'c', 'foo'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=2, b=3, c=4))

        self.assertTrue(if_all_in(['a', 'b'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=3, b=4, c=4))

    def test_if_any_in(self):
        a = dict(a=1, b=2, c=3)
        self.assertTrue(if_any_in(['a', 'b', 'c'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=2, b=3, c=4))

        self.assertTrue(if_any_in(['z', '45', 'a', 'b', 'c', 'foo'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=3, b=4, c=5))

        self.assertTrue(if_any_in(['a', 'b'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=4, b=5, c=5))

        self.assertFalse(if_any_in(['x', 'y'], a, lambda v: v + 1))
        self.assertDictEqual(a, dict(a=4, b=5, c=5))

    def test_try_cast(self):
        self.assertEqual(try_cast(int, '5'), 5)
        self.assertEqual(try_cast(float, '5.15'), 5.15)
        self.assertEqual(try_cast(float, '5.A15'), None)